from unittest.mock import patch

from atlasq import AtlasManager, AtlasQ
from atlasq.queryset.exceptions import AtlasIndexFieldError
from mongoengine import Document, ListField, StringField
from mongomock import command_cursor
from mongomock.command_cursor import CommandCursor
from tests.test_base import TestBaseCase


class MyDocument(Document):
    name = StringField(required=True)
    md5 = StringField(required=True)
    classification = StringField(required=True)
    related_threat = ListField(StringField())

    atlas = AtlasManager("test")


class TestQuerySet(TestBaseCase):
    def setUp(self) -> None:
        super(TestQuerySet, self).setUp()
        MyDocument.objects.all().delete()
        self.base = MyDocument.atlas
        self.obs = MyDocument(
            name="test.com",
            md5="d8cfbe774890e3b523ce584ce640a452",
            classification="domain",
            related_threat=["phishing"],
        )
        self.base.index.ensured = False

    def test_clone(self):
        self.base._count = True
        self.base.index.ensured = True
        qs = self.base.clone()
        self.assertEqual(qs._count, True)
        self.assertEqual(qs._aggrs_query, None)
        self.assertEqual(qs._search_result, None)
        self.assertNotEqual(qs.index, self.base.index)
        self.assertEqual(qs.index.index, self.base.index.index)
        self.assertEqual(qs.index.ensured, self.base.index.ensured)

    def test_ensure_index(self):
        with patch("atlasq.queryset.queryset.AtlasIndex.ensure_index_exists") as mock:
            self.base.ensure_index("user", "password", "group_id", "cluster_name")
            mock.assert_called_once_with(
                "user",
                "password",
                "group_id",
                "cluster_name",
                self.db_name,
                "my_document",
            )

    def test_count(self):
        r = self.base.count()
        self.assertEqual(r, 0)
        self.obs.save()
        r = self.base.count()
        self.assertEqual(r, 1)
        with patch(
            "mongomock.aggregate.process_pipeline",
            side_effect=[
                command_cursor.CommandCursor([{"meta": {"count": {"total": 0}}}]),
            ],
        ):
            r = self.base.filter(name="wrong.com").count()
        self.assertEqual(r, 0)

    def test_order_by(self):
        qs = self.base.order_by("-time")
        self.assertEqual(qs._aggrs[0], {"$sort": {"time": -1}})
        qs = self.base.order_by("+time")
        self.assertEqual(qs._aggrs[0], {"$sort": {"time": 1}})
        qs = self.base.order_by("time")
        self.assertEqual(qs._aggrs[0], {"$sort": {"time": 1}})
        qs = self.base.order_by("-time").filter(name="123")
        self.assertEqual(qs._aggrs[1], {"$sort": {"time": -1}})

    def test_only(self):
        qs = self.base.only("name").filter(name="123").order_by("-time")
        self.assertEqual(qs._get_projections(), [{"$project": {"name": 1}}])
        self.assertEqual(3, len(qs._aggrs))
        self.assertEqual(qs._aggrs[1], {"$project": {"name": 1}})

    def test_exclude(self):
        qs = self.base.exclude("name")
        self.assertEqual(qs._get_projections(), [{"$project": {"name": 0}}])

    def test_filter(self):
        qs = self.base.filter(name="test.com")
        self.assertEqual(
            qs._aggrs,
            [
                {
                    "$search": {
                        "index": "test",
                        "compound": {
                            "filter": [
                                {
                                    "text": {
                                        "query": "test.com",
                                        "path": "name",
                                    }
                                }
                            ]
                        },
                    }
                }
            ],
        )
        with patch(
            "mongomock.aggregate.process_pipeline",
            side_effect=[
                command_cursor.CommandCursor([{}]),
            ],
        ):
            self.assertEqual(len(qs), 0)
        self.obs.save()
        qs = self.base.filter(AtlasQ(name="test.com") & AtlasQ(classification="domain"))
        with patch(
            "mongomock.aggregate.process_pipeline",
            side_effect=[
                command_cursor.CommandCursor([{"_id": self.obs.id, "name": self.obs.name}]),
            ],
        ):
            self.assertEqual(len(qs), 1)

    def test_aggregate(self):
        cursor = self.base.aggregate({"$match": {"name": "test.com"}})
        self.assertIsInstance(cursor, CommandCursor)
        with self.assertRaises(StopIteration):
            next(cursor)
        self.obs.save()
        cursor = self.base.aggregate({"$match": {"name": "test.com"}})
        self.assertIsInstance(cursor, CommandCursor)
        obj = next(cursor)
        self.assertEqual(obj["_id"], self.obs.id)
        self.assertIn("name", obj)
        # exclude does not work with aggregate, even with mongoengine
        cursor = self.base.exclude("name").aggregate({"$match": {"name": "test.com"}})
        obj = next(cursor)
        self.assertIn("name", obj)

    def test_first(self):
        self.assertIsNone(self.base.first())
        self.obs.save()
        self.assertEqual(self.base.first().id, self.obs.id)
        with patch(
            "mongomock.aggregate.process_pipeline",
            side_effect=[
                command_cursor.CommandCursor([{"_id": self.obs.id, "name": self.obs.name}]),
            ],
        ):
            self.assertEqual(self.base.filter(name="test.com").first().name, self.obs.name)

    def test_get(self):
        with patch(
            "mongomock.aggregate.process_pipeline",
            side_effect=[
                command_cursor.CommandCursor([{}]),
            ],
        ):
            with self.assertRaises(MyDocument.DoesNotExist):
                self.base.get(name="test.com")
        self.obs.save()
        with patch(
            "mongomock.aggregate.process_pipeline",
            side_effect=[
                command_cursor.CommandCursor([{"_id": self.obs.id, "name": self.obs.name}]),
            ],
        ):
            self.assertEqual(self.base.get(name="test.com").name, self.obs.name)
        self.base.index.ensured = True
        with self.assertRaises(AtlasIndexFieldError):
            self.base.get(another_field="test.com")



from typing import Union

from atlasq.queryset.index import AtlasIndex
from atlasq.queryset.queryset import AtlasQuerySet
from mongoengine import QuerySetManager


class AtlasManager(QuerySetManager):
    """
    Manager for the Atlas class.
    """

    @property
    def default(self):
        if self._index:
            return AtlasQuerySet
        res = super().default
        return res

    def __init__(
        self,
        atlas_index: Union[str, None],
    ):
        super().__init__()
        self._index = AtlasIndex(atlas_index) if atlas_index else None

    def __get__(self, instance, owner):
        queryset = super().__get__(instance, owner)
        if isinstance(queryset, AtlasQuerySet):
            queryset.index = self._index
        return queryset



from .queryset.exceptions import AtlasIndexError, AtlasIndexFieldError
from .queryset.index import AtlasIndex
from .queryset.manager import AtlasManager
from .queryset.node import AtlasQ
from .queryset.queryset import AtlasQuerySet

__all__ = [
    "AtlasQ",
    "AtlasQuerySet",
    "AtlasManager",
    "AtlasIndex",
    "AtlasIndexFieldError",
    "AtlasIndexError",
]



import json

from atlasq import AtlasManager
from atlasq.queryset.index import AtlasIndex
from atlasq.queryset.node import AtlasQ
from atlasq.queryset.visitor import AtlasQueryCompilerVisitor
from mongoengine import Document, StringField
from tests.test_base import TestBaseCase


class MyDocument(Document):
    key = StringField()
    key2 = StringField()
    key3 = StringField()
    key4 = StringField()
    key5 = StringField()

    atlas = AtlasManager("test")


class TestAtlasQueryCompilerVisitor(TestBaseCase):
    def test_visit_combination_and_or(self):
        q1 = AtlasQ(key="value")

        q2 = AtlasQ(key2="value2")
        q3 = q1 | q2

        q4 = AtlasQ(key4="value4")

        q5 = q3 & q4

        q6 = AtlasQ(key5="value5")

        q7 = q5 | q6

        filters, *aggregations = q7.accept(AtlasQueryCompilerVisitor(MyDocument, AtlasIndex("test")))
        filters = filters["$search"]
        filters.pop("index")
        self.assertEqual([], aggregations)
        # print(json.dumps(filters, indent=4),)
        self.assertEqual(
            {
                "compound": {
                    "should": [
                        {
                            "compound": {
                                "filter": [
                                    {
                                        "compound": {
                                            "should": [
                                                {
                                                    "compound": {
                                                        "filter": [
                                                            {
                                                                "text": {
                                                                    "query": "value",
                                                                    "path": "key",
                                                                }
                                                            }
                                                        ]
                                                    }
                                                },
                                                {
                                                    "compound": {
                                                        "filter": [
                                                            {
                                                                "text": {
                                                                    "query": "value2",
                                                                    "path": "key2",
                                                                }
                                                            }
                                                        ]
                                                    }
                                                },
                                            ],
                                            "minimumShouldMatch": 1,
                                        }
                                    },
                                    {"text": {"query": "value4", "path": "key4"}},
                                ]
                            }
                        },
                        {"compound": {"filter": [{"text": {"query": "value5", "path": "key5"}}]}},
                    ],
                    "minimumShouldMatch": 1,
                }
            },
            filters,
            json.dumps(filters, indent=4),
        )

    def test_combination_and(self):
        q1 = AtlasQ(key="value", key2="value2")
        q2 = AtlasQ(key3="value3", key4="value4")
        q5 = q1 & q2
        filters, *aggregations = q5.accept(AtlasQueryCompilerVisitor(MyDocument, AtlasIndex("test")))
        filters = filters["$search"]
        filters.pop("index")
        self.assertEqual([], aggregations)
        self.assertEqual(
            {
                "compound": {
                    "filter": [
                        {"text": {"path": "key", "query": "value"}},
                        {"text": {"path": "key2", "query": "value2"}},
                        {"text": {"path": "key3", "query": "value3"}},
                        {"text": {"path": "key4", "query": "value4"}},
                    ],
                }
            },
            filters,
            json.dumps(filters, indent=4),
        )

    def test_atlas_q_or(self):
        q1 = AtlasQ(key="value", key2="value2")
        q2 = AtlasQ(key3="value3", key4="value4")
        q5 = q1 | q2
        filters, *aggregations = q5.accept(AtlasQueryCompilerVisitor(MyDocument, AtlasIndex("test")))
        filters = filters["$search"]
        filters.pop("index")
        self.assertEqual([], aggregations)
        self.assertEqual(
            {
                "compound": {
                    "should": [
                        {
                            "compound": {
                                "filter": [
                                    {"text": {"path": "key", "query": "value"}},
                                    {"text": {"path": "key2", "query": "value2"}},
                                ]
                            }
                        },
                        {
                            "compound": {
                                "filter": [
                                    {"text": {"path": "key3", "query": "value3"}},
                                    {"text": {"path": "key4", "query": "value4"}},
                                ]
                            }
                        },
                    ],
                    "minimumShouldMatch": 1,
                }
            },
            filters,
            json.dumps(filters, indent=4),
        )



import logging
from typing import Dict, List, Tuple, Union

from atlasq.queryset.visitor import AtlasQueryCompilerVisitor, AtlasSimplificationVisitor
from mongoengine import Q
from mongoengine.queryset.visitor import QCombination

logger = logging.getLogger(__name__)


class AtlasQ(Q):
    @property
    def operation(self):
        return self.AND

    def to_query(self, document) -> List[Dict]:  # pylint: disable=arguments-differ
        from atlasq import AtlasQuerySet

        qs = getattr(document, "atlas", None)
        if qs is None:
            raise ValueError("Document must set `atlas` to an AtlasManager")
        if not isinstance(qs, AtlasQuerySet):
            return super().to_query(document)
        atlas_index = qs.index
        logger.debug(f"to_query {self.__class__.__name__} {document}")
        query = self.accept(AtlasSimplificationVisitor())
        query = query.accept(AtlasQueryCompilerVisitor(document, atlas_index))
        return query

    def _combine(self, other, operation) -> Union["AtlasQ", "AtlasQCombination"]:
        logger.debug(f"_combine {self.__class__.__name__} {other}, {operation}")

        result = super()._combine(other, operation)
        if isinstance(result, QCombination):
            return AtlasQCombination(result.operation, result.children)
        return AtlasQ(**result.query)


class AtlasQCombination(QCombination):
    def __bool__(self):
        return any(bool(child) for child in self.children)

    def _combine(self, other, operation):
        logger.debug(f"_combine {self.__class__.__name__} {other}, {operation}")
        result = super()._combine(other, operation)
        return AtlasQCombination(result.operation, result.children)

    def to_query(self, document) -> Tuple[Dict, List[Dict]]:  # pylint: disable=arguments-differ
        from atlasq import AtlasQuerySet

        qs = getattr(document, "atlas", None)
        if qs is None:
            raise ValueError("Document must set `atlas` to an AtlasManager")
        if not isinstance(qs, AtlasQuerySet):
            return super().to_query(document)
        atlas_index = qs.index
        logger.debug(f"to_query {self.__class__.__name__} {document}")
        query = self.accept(AtlasSimplificationVisitor())
        query = query.accept(AtlasQueryCompilerVisitor(document, atlas_index))
        return query

    def accept(self, visitor):
        logger.debug(f"accept {self.__class__.__name__} {visitor}")
        return super().accept(visitor)



class AtlasError(Exception):
    """Base class for all AtlasQ exceptions."""


class AtlasIndexFieldError(AtlasError):
    pass


class AtlasIndexError(AtlasError):
    pass


class AtlasFieldError(AtlasError):
    pass



import fnmatch
from enum import Enum
from logging import getLogger
from typing import Dict, List

import requests
from atlasq.queryset.exceptions import AtlasIndexError, AtlasIndexFieldError
from requests.auth import HTTPDigestAuth

logger = getLogger(__name__)

ATLAS_BASE_URL = "https://cloud.mongodb.com/api/atlas/v1.0"
TEXT_INDEXES_ENDPOINT = ATLAS_BASE_URL + "/groups/{GROUP_ID}/clusters/{CLUSTER_NAME}/fts/indexes"

LIST_TEXT_INDEXES_ENDPOINT = TEXT_INDEXES_ENDPOINT + "/{DATABASE_NAME}/{COLLECTION_NAME}"


class AtlasIndexType(Enum):
    DOCUMENT = "document"
    EMBEDDED_DOCUMENT = "embeddedDocuments"
    STRING = "string"
    NUMBER = "number"
    BOOLEAN = "boolean"
    DATE = "date"
    OBJECT_ID = "objectId"

    @classmethod
    def values(cls) -> List[str]:
        return [e.value for e in cls]


class AtlasIndex:

    fields_to_copy = ["ensured", "_indexed_fields"]

    def __init__(self, index_name: str):
        self._indexed_fields: Dict[str, str] = {}
        self.ensured: bool = False
        self._index: str = index_name

    def __copy__(self):
        res = AtlasIndex(self._index)
        for field in self.fields_to_copy:
            setattr(res, field, getattr(self, field))
        return res

    @property
    def index(self) -> str:
        return self._index

    @index.setter
    def index(self, index: str):
        self._index = index
        self.ensured = False

    # pylint: disable=too-many-arguments
    def upload_index(
        self,
        data: Dict,
        user: str,
        password: str,
        group_id: str,
        cluster_name: str,
    ):
        if not self.index:
            raise AtlasIndexError("No index defined")

        if not isinstance(data, dict):
            raise AtlasIndexError("The index should be a dictionary")
        if "mappings" not in data:
            raise AtlasIndexError("There is no 'mappings' in the index")
        if not isinstance(data["mappings"], dict):
            raise AtlasIndexError("The mappings keyword should be a dictionary")
        if "fields" not in data["mappings"]:
            raise AtlasIndexError("There is no 'mappings' in the index")
        if not isinstance(data["mappings"]["fields"], dict):
            raise AtlasIndexError("The fields keyword should be a dictionary")

        if any(id_keyword in data["mappings"]["fields"] for id_keyword in ["id", "pk"]):
            if "_id" not in data["mappings"]["fields"]:
                data["mappings"]["fields"]["_id"] = {"type": AtlasIndexType.OBJECT_ID.value}

        url = TEXT_INDEXES_ENDPOINT.format(
            GROUP_ID=group_id,
            CLUSTER_NAME=cluster_name,
        )
        response = requests.post(
            url,
            json=data,
            auth=HTTPDigestAuth(user, password),
            headers={"Content-Type": "application/json"},
        )
        response.raise_for_status()

    # pylint: disable=too-many-arguments
    def ensure_index_exists(
        self,
        user: str,
        password: str,
        group_id: str,
        cluster_name: str,
        db_name: str,
        collection_name: str,
    ):
        if not self.index:
            raise AtlasIndexError("No index defined")
        url = LIST_TEXT_INDEXES_ENDPOINT.format(
            GROUP_ID=group_id,
            CLUSTER_NAME=cluster_name,
            DATABASE_NAME=db_name,
            COLLECTION_NAME=collection_name,
        )
        response = requests.get(url, auth=HTTPDigestAuth(user, password))
        response.raise_for_status()
        index_results = response.json()
        self._indexed_fields.clear()
        for index_result in index_results:
            if index_result["name"] == self.index:
                self._set_indexed_from_mappings(index_result)
                self.ensured = True
                break
        else:
            self.ensured = False
        return self.ensured

    def _set_indexed_fields(self, index_result: Dict, base_field: str = ""):
        lucene_type = index_result["type"]
        if lucene_type in [
            AtlasIndexType.DOCUMENT.value,
            AtlasIndexType.EMBEDDED_DOCUMENT.value,
        ]:
            if not index_result.get("dynamic", False):
                for field, value in index_result.get("fields", {}).items():
                    field = f"{base_field}.{field}" if base_field else field
                    self._set_indexed_fields(value, base_field=field)
            else:
                self._indexed_fields[f"{base_field}.*" if base_field else "*"] = ""
        if base_field:
            if lucene_type not in AtlasIndexType.values():
                logger.warning(f"Lucene type {lucene_type} not configured")
            self._indexed_fields[base_field] = lucene_type

    def _set_indexed_from_mappings(self, index_result: Dict):
        mappings = index_result["mappings"]
        mappings["type"] = AtlasIndexType.DOCUMENT.value
        self._set_indexed_fields(mappings)
        logger.debug(self._indexed_fields)

    def ensure_keyword_is_indexed(self, keyword: str):
        if not self.ensured:
            raise AtlasIndexError("Index not ensured")
        return any(fnmatch.fnmatch(keyword, field) for field in self._indexed_fields)

    def get_type_from_keyword(self, keyword) -> str:
        if not self.ensured:
            raise AtlasIndexError("Index not ensured")
        if keyword in self._indexed_fields:
            return self._indexed_fields[keyword]
        raise AtlasIndexFieldError(f"Keyword {keyword} not present in index")



from unittest import TestCase

from mongoengine import connect, disconnect


class TestBaseCase(TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.db_name = "mongoenginetest"
        connect(cls.db_name, host="mongomock://localhost")

    @classmethod
    def tearDownClass(cls) -> None:
        disconnect()



import logging
from typing import Dict, List, Tuple

from atlasq.queryset.index import AtlasIndex
from mongoengine import Q
from mongoengine.queryset.visitor import QueryCompilerVisitor, SimplificationVisitor

logger = logging.getLogger(__name__)


class AtlasSimplificationVisitor(SimplificationVisitor):
    def visit_combination(self, combination):
        from atlasq.queryset.node import AtlasQ

        logger.debug(f"visit_combination {self.__class__.__name__} {combination}")
        result = super().visit_combination(combination)
        if isinstance(result, Q) and not isinstance(result, AtlasQ):
            return AtlasQ(**result.query)
        return result

    def visit_query(self, query):
        logger.debug(f"visit_query {self.__class__.__name__} {query}")
        return super().visit_query(query)


class AtlasQueryCompilerVisitor(QueryCompilerVisitor):
    def __init__(self, document, atlas_index: AtlasIndex):
        super().__init__(document)
        self.atlas_index = atlas_index

    def _visit_combination_and(self, combination) -> Tuple[Dict, List[Dict]]:
        affirmatives = []
        negatives = []
        aggregations = []
        for child in combination.children:
            filters, *child_aggregations = child
            try:
                filters = filters["$search"]
            except KeyError:
                # in case we return just aggregations, the filters should be empty
                child_aggregations = [filters] + child_aggregations
                filters = {}
            if "compound" in filters:
                if "filter" in filters["compound"]:
                    affirmatives.extend(filters["compound"]["filter"])
                if "mustNot" in filters["compound"]:
                    negatives.extend(filters["compound"]["mustNot"])
                if "should" in filters["compound"]:
                    filters.pop("index")
                    affirmatives.append(filters)
            aggregations.extend(child_aggregations)
        result = {"compound": {}}
        if affirmatives:
            result["compound"]["filter"] = affirmatives
        if negatives:
            result["compound"]["mustNot"] = negatives
        return result, aggregations

    def _visit_combination_or(self, combination) -> Tuple[Dict, List[Dict]]:
        aggregations = []
        children_results = []
        for child in combination.children:
            filters, *child_aggregations = child
            try:
                filters = filters["$search"]
            except KeyError:
                # in case we return just aggregations, the filters should be empty
                child_aggregations = [filters] + child_aggregations
            else:
                filters.pop("index")
                children_results.append(filters)
            aggregations.extend(child_aggregations)
        return {"compound": {"should": children_results, "minimumShouldMatch": 1}}, aggregations

    def visit_combination(self, combination) -> List[Dict]:
        if combination.operation == combination.AND:
            filters, aggregations = self._visit_combination_and(combination)
        else:
            filters, aggregations = self._visit_combination_or(combination)
        filters["index"] = self.atlas_index.index
        result = [{"$search": filters}] + aggregations
        return result

    def visit_query(self, query) -> List[Dict]:
        from atlasq.queryset.transform import AtlasTransform

        affirmative, negative, aggregations = AtlasTransform(query.query, self.atlas_index).transform()
        filters = {}
        if affirmative:
            filters.setdefault("compound", {})["filter"] = affirmative
        if negative:
            filters.setdefault("compound", {})["mustNot"] = negative
        result = []
        if filters:
            filters["index"] = self.atlas_index.index
            result += [{"$search": filters}]
        result += aggregations
        return result



from unittest.mock import patch

from atlasq.queryset.index import AtlasIndex
from requests import HTTPError
from tests.test_base import TestBaseCase


class MockResponse:
    def __init__(self, json_data, status_code):
        self.json_data = json_data
        self.status_code = status_code

    def json(self):
        return self.json_data

    def raise_for_status(self):
        if not 200 <= self.status_code <= 299:
            raise HTTPError(self.status_code)


class TestIndex(TestBaseCase):
    def test_ensure_keyword_is_indexed(self):
        index = AtlasIndex("myindex")
        index._indexed_fields = ["field1", "field2.*"]
        index.ensured = True
        self.assertTrue(index.ensure_keyword_is_indexed("field1"))
        self.assertFalse(index.ensure_keyword_is_indexed("field2"))
        self.assertTrue(index.ensure_keyword_is_indexed("field2.field3"))

    def test_set_indexed_from_mappings(self):
        index = AtlasIndex("myindex")
        index._set_indexed_from_mappings(
            {
                "mappings": {
                    "dynamic": False,
                    "fields": {
                        "field1": {"type": "document", "dynamic": True},
                        "field2": {
                            "dynamic": False,
                            "fields": {
                                "field": {"indexOptions": "docs", "type": "string"},
                                "field4": {
                                    "fields": {
                                        "field5": {
                                            "indexOptions": "docs",
                                            "type": "string",
                                        },
                                        "field6": {
                                            "indexOptions": "docs",
                                            "type": "string",
                                        },
                                    },
                                    "type": "document",
                                },
                            },
                            "type": "document",
                        },
                    },
                }
            }
        )
        self.assertCountEqual(
            index._indexed_fields.keys(),
            [
                "field1",
                "field1.*",
                "field2",
                "field2.field",
                "field2.field4",
                "field2.field4.field5",
                "field2.field4.field6",
            ],
        )

    def test_set_indexed_fields(self):
        index = AtlasIndex("myindex")
        index._set_indexed_fields(
            {
                "type": "document",
                "dynamic": False,
                "fields": {
                    "field1": {"indexOptions": "docs", "type": "string"},
                    "field2": {
                        "fields": {
                            "field3": {"indexOptions": "docs", "type": "string"},
                            "field4": {"indexOptions": "docs", "type": "string"},
                        },
                        "type": "document",
                    },
                },
            }
        )
        self.assertCountEqual(
            index._indexed_fields,
            ["field1", "field2", "field2.field3", "field2.field4"],
        )
        index._indexed_fields.clear()
        index._set_indexed_fields(
            {
                "type": "document",
                "dynamic": True,
            }
        )
        self.assertCountEqual(index._indexed_fields, ["*"])

    def test_ensure_index_exists(self):
        index = AtlasIndex("myindex")
        self.assertFalse(index.ensured)
        with patch(
            "requests.get",
            return_value=MockResponse(
                [
                    {
                        "collectionName": "COLL",
                        "database": "DB",
                        "indexID": "ID",
                        "mappings": {
                            "dynamic": False,
                            "fields": {
                                "field1": {"type": "boolean"},
                                "field2": {"indexOptions": "docs", "type": "string"},
                            },
                        },
                        "name": "myindex",
                        "status": "STEADY",
                        "synonyms": [],
                    }
                ],
                200,
            ),
        ):
            result = index.ensure_index_exists("user", "password", "project", "cluster", "db", "collection")
            self.assertTrue(result)
            self.assertTrue(index.ensured)
            self.assertCountEqual(index._indexed_fields, ["field1", "field2"])
            self.assertTrue(index.ensure_keyword_is_indexed("field1"))
            self.assertFalse(index.ensure_keyword_is_indexed("field3"))
        with patch(
            "requests.get",
            return_value=MockResponse(
                [
                    {
                        "collectionName": "COLL",
                        "database": "DB",
                        "indexID": "ID",
                        "mappings": {
                            "dynamic": False,
                            "fields": {
                                "field1": {"type": "boolean"},
                                "field2": {"indexOptions": "docs", "type": "string"},
                            },
                        },
                        "name": "anotherindex",
                        "status": "STEADY",
                        "synonyms": [],
                    }
                ],
                200,
            ),
        ):
            result = index.ensure_index_exists("user", "password", "project", "cluster", "db", "collection")
            self.assertFalse(result)
            self.assertFalse(index.ensured)
            self.assertCountEqual(index._indexed_fields, [])
        with patch(
            "requests.get",
            return_value=MockResponse(
                {
                    "detail": "Current user is not authorized to perform this action.",
                    "error": 401,
                    "errorCode": "USER_UNAUTHORIZED",
                    "parameters": [],
                    "reason": "Unauthorized",
                },
                401,
            ),
        ):
            with self.assertRaises(HTTPError):
                index.ensure_index_exists("user", "password", "group", "cluster", "db", "collection")



from atlasq.queryset.manager import AtlasManager
from atlasq.queryset.queryset import AtlasQuerySet
from mongoengine import Document, QuerySet, fields
from tests.test_base import TestBaseCase


class TestManager(TestBaseCase):
    def test_with_index(self):
        class MyDocument(Document):
            name = fields.StringField(required=True)
            atlas = AtlasManager("myindex")

        self.assertIsInstance(MyDocument.atlas, AtlasQuerySet)
        self.assertEqual(MyDocument.atlas.index.index, "myindex")
        # self.assertIsNotNone(MyDocument.atlas.cache)

    def test_no_index(self):
        class MyDocument(Document):
            name = fields.StringField(required=True)
            atlas = AtlasManager(None)

        self.assertIsInstance(MyDocument.atlas, QuerySet)
        with self.assertRaises(AssertionError):
            self.assertIsInstance(MyDocument.atlas, AtlasQuerySet)



from atlasq import AtlasManager
from atlasq.queryset.node import AtlasQ, AtlasQCombination
from mongoengine import Document, fields
from tests.test_base import TestBaseCase


class TestAtlasQ(TestBaseCase):
    def test__combine_empty(self):
        q = AtlasQ(field=3)
        q2 = AtlasQ()
        self.assertEqual(q._combine(q2, q.AND), q)

    def test_combine(self):
        q = AtlasQ(field=3)
        q2 = AtlasQ(field2=4)
        self.assertEqual(q._combine(q2, q.AND), AtlasQCombination(q.AND, [q, q2]))
        self.assertEqual(q._combine(q2, q.OR), AtlasQCombination(q.OR, [q, q2]))

    def test_bool(self):
        self.assertFalse(bool(AtlasQ()))
        self.assertTrue(bool(AtlasQ(f=1)))

    def test_to_query(self):
        class MyDocument(Document):
            name = fields.StringField()

            atlas = AtlasManager("test")

        q = AtlasQ(name="test")
        res = q.to_query(MyDocument)
        self.assertIsInstance(res, list)
        self.assertEqual(1, len(res))
        filters, *aggregations = res
        self.assertIsInstance(filters, dict)
        self.assertIn("compound", filters["$search"])
        self.assertIsInstance(aggregations, list)
        self.assertEqual(0, len(aggregations))


class TestAtlasQCombination(TestBaseCase):
    def test_bool(self):
        self.assertFalse(bool(AtlasQCombination(AtlasQCombination.AND, [AtlasQ()])))
        self.assertTrue(bool(AtlasQCombination(AtlasQCombination.AND, [AtlasQ(f=1)])))

    def test__combine(self):
        q1 = AtlasQCombination(AtlasQCombination.AND, [AtlasQ(field=3), AtlasQ(field2=4)])
        q2 = AtlasQCombination(AtlasQCombination.AND, [AtlasQ(field3=3), AtlasQ(field4=4)])
        self.assertIsInstance(q1._combine(q2, q1.AND), AtlasQCombination)



import datetime
import json
from unittest import expectedFailure

from atlasq.queryset.exceptions import AtlasFieldError, AtlasIndexFieldError
from atlasq.queryset.index import AtlasIndex
from atlasq.queryset.node import AtlasQ
from atlasq.queryset.transform import AtlasTransform
from bson import ObjectId
from mongoengine import Document, fields
from tests.test_base import TestBaseCase


class TestTransformSteps(TestBaseCase):
    def test_convert_type_keyword_list_datetime(self):
        nnow = datetime.datetime.now()
        tomorrow = nnow + datetime.timedelta(days=1)
        q = AtlasQ(f=[nnow, tomorrow])
        t = AtlasTransform(q.query, AtlasIndex("test"))
        result = t._auto_convert_type_to_keyword("f", [nnow, tomorrow])
        self.assertEqual(
            result,
            {
                "compound": {
                    "should": [
                        {
                            "equals": {
                                "path": "f",
                                "value": nnow,
                            }
                        },
                        {
                            "equals": {
                                "path": "f",
                                "value": tomorrow,
                            }
                        },
                    ],
                    "minimumShouldMatch": 1,
                }
            },
        )

    def test_convert_type_keyword_datetime(self):
        nnow = datetime.datetime.now()
        q = AtlasQ(f=nnow)
        t = AtlasTransform(q.query, AtlasIndex("test"))
        result = t._auto_convert_type_to_keyword("f", nnow)
        self.assertEqual(
            result,
            {
                "equals": {
                    "path": "f",
                    "value": nnow,
                }
            },
        )

    def test_convert_type_keyword_list_bool(self):
        q = AtlasQ(f=[True, False])
        t = AtlasTransform(q.query, AtlasIndex("test"))
        result = t._auto_convert_type_to_keyword("f", [True, False])
        self.assertEqual(
            result,
            {
                "compound": {
                    "should": [
                        {
                            "equals": {
                                "path": "f",
                                "value": True,
                            }
                        },
                        {
                            "equals": {
                                "path": "f",
                                "value": False,
                            }
                        },
                    ],
                    "minimumShouldMatch": 1,
                }
            },
        )

    def test_convert_type_keyword_bool(self):
        q = AtlasQ(f=True)
        t = AtlasTransform(q.query, AtlasIndex("test"))
        result = t._auto_convert_type_to_keyword("f", True)
        self.assertEqual(
            result,
            {
                "equals": {
                    "path": "f",
                    "value": True,
                }
            },
        )

    def test_convert_type_to_keyword_objectId(self):
        q = AtlasQ(f=ObjectId("5e45de3dd2bfea029b68cce2"))
        t = AtlasTransform(q.query, AtlasIndex("test"))
        result = t._auto_convert_type_to_keyword("f", ObjectId("5e45de3dd2bfea029b68cce2"))
        self.assertEqual(
            result,
            {
                "equals": {
                    "path": "f",
                    "value": ObjectId("5e45de3dd2bfea029b68cce2"),
                }
            },
        )

    def test_convert_type_to_keyword_list_objectId(self):
        q = AtlasQ(f=[ObjectId("5e45de3dd2bfea029b68cce2"), ObjectId("5e45de3dd2bfea029b68cce3")])
        t = AtlasTransform(q.query, AtlasIndex("test"))
        result = t._auto_convert_type_to_keyword("f", [ObjectId("5e45de3dd2bfea029b68cce2"), ObjectId("5e45de3dd2bfea029b68cce3")])
        self.assertEqual(
            result,
            {
                "compound": {
                    "should": [
                        {
                            "equals": {
                                "path": "f",
                                "value": ObjectId("5e45de3dd2bfea029b68cce2"),
                            }
                        },
                        {
                            "equals": {
                                "path": "f",
                                "value": ObjectId("5e45de3dd2bfea029b68cce3"),
                            }
                        },
                    ],
                    "minimumShouldMatch": 1,
                }
            },
        )

    def test_all(self):
        q = AtlasQ(f__all=["one", "two"])
        t = AtlasTransform(q.query, AtlasIndex("test"))
        res = t._all("f", ["one", "two"])
        self.assertEqual(
            res,
            {
                "compound": {
                    "must": [
                        {"text": {"query": "one", "path": "f"}},
                        {"text": {"query": "two", "path": "f"}},
                    ]
                }
            },
        )

    def test_merge_embedded_documents_to_not_embedded(self):
        list_of_objs = [
            {
                "text": {
                    "query": "aaa",
                    "path": "field.field2.field3",
                }
            }
        ]
        obj = {
            "embeddedDocument": {
                "path": "field",
                "operator": {"compound": {"must": [{"text": {"query": "bbb", "path": "field.field4"}}]}},
            }
        }
        AtlasTransform.merge_embedded_documents(obj, list_of_objs)

    def test_merge_embedded_documents_multi_level_different_level(self):
        obj = {
            "embeddedDocument": {
                "path": "field",
                "operator": {
                    "compound": {
                        "must": [
                            {
                                "embeddedDocument": {
                                    "path": "field.field2",
                                    "operator": {
                                        "compound": {
                                            "must": [
                                                {
                                                    "text": {
                                                        "query": "aaa",
                                                        "path": "field.field2.field3",
                                                    }
                                                }
                                            ]
                                        }
                                    },
                                }
                            }
                        ]
                    }
                },
            }
        }

        list_of_objs = [
            {
                "embeddedDocument": {
                    "path": "field",
                    "operator": {"compound": {"must": [{"text": {"query": "bbb", "path": "field.field4"}}]}},
                }
            }
        ]
        result = AtlasTransform.merge_embedded_documents(obj, list_of_objs)
        self.assertIsInstance(result, list)
        self.assertEqual(1, len(result))
        self.assertEqual(
            result[0],
            {
                "embeddedDocument": {
                    "path": "field",
                    "operator": {
                        "compound": {
                            "must": [
                                {"text": {"query": "bbb", "path": "field.field4"}},
                                {
                                    "embeddedDocument": {
                                        "path": "field.field2",
                                        "operator": {
                                            "compound": {
                                                "must": [
                                                    {
                                                        "text": {
                                                            "query": "aaa",
                                                            "path": "field.field2.field3",
                                                        }
                                                    }
                                                ]
                                            }
                                        },
                                    }
                                },
                            ]
                        }
                    },
                }
            },
        )

    @expectedFailure
    def test_merge_embedded_documents_multi_level_same_level(self):
        obj = {
            "embeddedDocument": {
                "path": "field",
                "operator": {
                    "compound": {
                        "must": [
                            {
                                "embeddedDocument": {
                                    "path": "field.field2",
                                    "operator": {
                                        "compound": {
                                            "must": [
                                                {
                                                    "text": {
                                                        "query": "aaa",
                                                        "path": "field.field2.field3",
                                                    }
                                                }
                                            ]
                                        }
                                    },
                                }
                            }
                        ]
                    }
                },
            }
        }
        list_of_objs = [
            {
                "embeddedDocument": {
                    "path": "field",
                    "operator": {
                        "compound": {
                            "must": [
                                {
                                    "embeddedDocument": {
                                        "path": "field.field2",
                                        "operator": {
                                            "compound": {
                                                "must": [
                                                    {
                                                        "text": {
                                                            "query": "aaa",
                                                            "path": "field.field2.field4",
                                                        }
                                                    }
                                                ]
                                            }
                                        },
                                    }
                                }
                            ]
                        }
                    },
                }
            }
        ]
        result = AtlasTransform.merge_embedded_documents(obj, list_of_objs)
        self.assertIsInstance(result, list)
        self.assertEqual(1, len(result))
        self.assertEqual(
            result[0],
            {
                "embeddedDocument": {
                    "path": "field",
                    "operator": {
                        "compound": {
                            "must": [
                                {
                                    "embeddedDocument": {
                                        "path": "field.field2",
                                        "operator": {
                                            "compound": {
                                                "must": [
                                                    {
                                                        "text": {
                                                            "query": "aaa",
                                                            "path": "field.field2.field3",
                                                        }
                                                    },
                                                    {
                                                        "text": {
                                                            "query": "aaa",
                                                            "path": "field.field2.field4",
                                                        }
                                                    },
                                                ]
                                            }
                                        },
                                    }
                                },
                            ]
                        }
                    },
                }
            },
        )

    def test_merge_embedded_documents(self):
        obj = {
            "embeddedDocument": {
                "path": "field",
                "operator": {"compound": {"must": [{"text": {"query": "aaa", "path": "field.field2"}}]}},
            }
        }
        list_of_objs = []

        result = AtlasTransform.merge_embedded_documents(obj, list_of_objs)
        self.assertEqual(0, len(list_of_objs))
        self.assertIsInstance(result, list)
        self.assertEqual(1, len(result))
        self.assertEqual(obj, result[0])

        list_of_objs = [
            {
                "embeddedDocument": {
                    "path": "field",
                    "operator": {"compound": {"must": [{"text": {"query": "bbb", "path": "field.field3"}}]}},
                }
            }
        ]
        result = AtlasTransform.merge_embedded_documents(obj, list_of_objs)
        self.assertEqual(1, len(result))
        self.assertEqual(
            result[0],
            {
                "embeddedDocument": {
                    "path": "field",
                    "operator": {
                        "compound": {
                            "must": [
                                {"text": {"query": "bbb", "path": "field.field3"}},
                                {"text": {"query": "aaa", "path": "field.field2"}},
                            ]
                        }
                    },
                }
            },
        )

        list_of_objs = [
            {
                "embeddedDocument": {
                    "path": "field",
                    "operator": {"compound": {"mustNot": [{"text": {"query": "bbb", "path": "field.field3"}}]}},
                }
            }
        ]
        result = AtlasTransform.merge_embedded_documents(obj, list_of_objs)
        self.assertEqual(1, len(result))
        self.assertEqual(
            result[0],
            {
                "embeddedDocument": {
                    "path": "field",
                    "operator": {
                        "compound": {
                            "must": [{"text": {"query": "aaa", "path": "field.field2"}}],
                            "mustNot": [{"text": {"query": "bbb", "path": "field.field3"}}],
                        }
                    },
                }
            },
        )

    def test__multiple_check_single_embedded_document(self):
        index = AtlasIndex("test")
        index.ensured = True
        index._indexed_fields = {
            "field": "embeddedDocuments",
            "field.field2": "string",
            "field.field3": "string",
        }
        q = AtlasQ(field__field2="aaa", field__field3__ne="bbb")
        try:
            result = AtlasTransform(q.query, index).transform()
        except AtlasIndexFieldError as e:
            self.fail(e)
        else:
            print(result[0][0])
            self.assertEqual(
                result[0][0],
                {
                    "embeddedDocument": {
                        "path": "field",
                        "operator": {
                            "compound": {
                                "must": [{"text": {"query": "aaa", "path": "field.field2"}}],
                                "mustNot": [{"text": {"query": "bbb", "path": "field.field3"}}],
                            }
                        },
                    }
                },
            )

    def test__embedded_document(self):
        index = AtlasIndex("test")
        index.ensured = True
        index._indexed_fields = {"field": "embeddedDocuments", "field.field2": "string"}
        q = AtlasQ(field__field2="aaa")
        try:
            result = AtlasTransform(q.query, index).transform()
        except AtlasIndexFieldError as e:
            self.fail(e)
        else:
            self.assertEqual(1, len(result[0]))
            self.assertEqual(
                result[0][0],
                {
                    "embeddedDocument": {
                        "path": "field",
                        "operator": {
                            "compound": {
                                "must": [
                                    {
                                        "text": {
                                            "query": "aaa",
                                            "path": "field.field2",
                                        },
                                    }
                                ]
                            }
                        },
                    }
                },
            )

        index._indexed_fields = {"field": "document", "field.field2": "string"}
        q = AtlasQ(field__field2="aaa")
        try:
            result = AtlasTransform(q.query, index).transform()
        except AtlasIndexFieldError as e:
            self.fail(e)
        else:
            self.assertEqual(result[0][0], {"text": {"query": "aaa", "path": "field.field2"}})

        index._indexed_fields = {
            "field": "document",
            "field.field2": "document",
            "field.field2.field3": "string",
        }
        q = AtlasQ(field__field2__field3="aaa")
        try:
            result = AtlasTransform(q.query, index).transform()
        except AtlasIndexFieldError as e:
            self.fail(e)
        else:
            self.assertEqual(result[0][0], {"text": {"query": "aaa", "path": "field.field2.field3"}})

        index._indexed_fields = {
            "field": "document",
            "field.field2": "embeddedDocuments",
            "field.field2.field3": "string",
        }
        q = AtlasQ(field__field2__field3="aaa")
        try:
            result = AtlasTransform(q.query, index).transform()
        except AtlasIndexFieldError as e:
            self.fail(e)
        else:
            self.assertEqual(result[0][0], {"text": {"query": "aaa", "path": "field.field2.field3"}})

        index._indexed_fields = {
            "field": "embeddedDocuments",
            "field.field2": "embeddedDocuments",
            "field.field2.field3": "string",
        }
        q = AtlasQ(field__field2__field3="aaa")
        try:
            result = AtlasTransform(q.query, index).transform()
        except AtlasIndexFieldError as e:
            self.fail(e)
        else:
            self.assertEqual(
                result[0][0],
                {
                    "embeddedDocument": {
                        "path": "field",
                        "operator": {
                            "compound": {
                                "must": [
                                    {
                                        "embeddedDocument": {
                                            "path": "field.field2",
                                            "operator": {
                                                "compound": {
                                                    "must": [
                                                        {
                                                            "text": {
                                                                "query": "aaa",
                                                                "path": "field.field2.field3",
                                                            }
                                                        },
                                                    ]
                                                }
                                            },
                                        }
                                    },
                                ]
                            }
                        },
                    }
                },
            )

        index._indexed_fields = {
            "field": "embeddedDocuments",
            "field.field2": "document",
            "field.field2.field3": "string",
        }
        q = AtlasQ(field__field2__field3="aaa")
        try:
            result = AtlasTransform(q.query, index).transform()
        except AtlasIndexFieldError as e:
            self.fail(e)
        else:
            self.assertEqual(
                result[0][0],
                {
                    "embeddedDocument": {
                        "path": "field",
                        "operator": {
                            "compound": {
                                "must": [
                                    {
                                        "text": {
                                            "query": "aaa",
                                            "path": "field.field2.field3",
                                        }
                                    },
                                ]
                            }
                        },
                    }
                },
            )

    def test__ensure_keyword(self):

        index = AtlasIndex("test")
        index._indexed_fields = {"field": "string"}
        q = AtlasQ(field="aaa")
        try:
            AtlasTransform(q.query, index).transform()
        except AtlasIndexFieldError as e:
            self.fail(e)
        q = AtlasQ(field2="aaa")
        try:
            AtlasTransform(q.query, index).transform()
        except AtlasIndexFieldError as e:
            self.fail(e)

        index.ensured = True

        with self.assertRaises(AtlasIndexFieldError):
            AtlasTransform(q.query, index).transform()

        q = AtlasQ(field__field2="bbb")

        with self.assertRaises(AtlasIndexFieldError):
            AtlasTransform(q.query, index).transform()

        index._indexed_fields["field.field2"] = "string"
        try:
            AtlasTransform(q.query, index).transform()
        except AtlasIndexFieldError as e:
            self.fail(e)

    def test__regex(self):
        q = AtlasQ(f__regex=".*")
        t = AtlasTransform(q.query, AtlasIndex("test"))
        res = t._regex("f", ".*")
        self.assertIn("regex", res)
        self.assertIn("query", res["regex"])
        self.assertIn(".*", res["regex"]["query"])
        self.assertIn("path", res["regex"])
        self.assertIn("f", res["regex"]["path"])

    def test__queryset_value(self):
        class MyDoc(Document):
            field = fields.StringField()

        MyDoc.objects.create(field="aaa")
        objects = MyDoc.objects.all().values_list("field")

        q = AtlasQ(f=objects)
        positive, negative, aggregations = AtlasTransform(q.query, AtlasIndex("test")).transform()

        self.assertEqual(positive, [{"text": {"query": ["aaa"], "path": "f"}}])
        self.assertEqual(negative, [])
        self.assertEqual(aggregations, [])

    def test__exists(self):
        q = AtlasQ(f=3)
        t = AtlasTransform(q.query, AtlasIndex("test"))
        res = t._exists("field")
        self.assertIn("exists", res)
        self.assertIn("path", res["exists"])
        self.assertEqual(res["exists"]["path"], "field")

    def test__range_date_valid(self):
        q = AtlasQ(f=3)
        t = AtlasTransform(q.query, AtlasIndex("test"))
        time = datetime.datetime.now()
        res = t._range("field", time, ["lte"])
        self.assertEqual(
            res,
            {
                "range": {
                    "path": "field",
                    "lte": time.replace(microsecond=0),
                }
            },
        )

    def test__range_integer_valid(self):
        q = AtlasQ(f=3)
        t = AtlasTransform(q.query, AtlasIndex("test"))
        res = t._range("field", 3, ["lte"])
        self.assertEqual(res, {"range": {"path": "field", "lte": 3}})

    def test__range_date_invalid(self):
        q = AtlasQ(f=3)
        t = AtlasTransform(q.query, AtlasIndex("test"))
        time = datetime.datetime.now()
        with self.assertRaises(AtlasFieldError):
            t._range("field", time, ["wat"])

    def test__range_none(self):
        q = AtlasQ(f=3)
        t = AtlasTransform(q.query, AtlasIndex("test"))
        with self.assertRaises(AtlasFieldError):
            t._range("field", None, ["lte"])

    def test__range_string(self):
        q = AtlasQ(f=3)
        t = AtlasTransform(q.query, AtlasIndex("test"))
        with self.assertRaises(AtlasFieldError):
            t._range("field", "3", ["lte"])

    def test__equals_string(self):
        q = AtlasQ(f=3)
        t = AtlasTransform(q.query, AtlasIndex("test"))
        with self.assertRaises(AtlasFieldError):
            t._equals("field", "aaa")

    def test__equals_bool(self):
        q = AtlasQ(f=3)
        t = AtlasTransform(q.query, AtlasIndex("test"))
        res = t._equals("field", True)
        self.assertIn("equals", res)
        self.assertIn("path", res["equals"])
        self.assertIn("value", res["equals"])
        self.assertEqual(res["equals"]["path"], "field")
        self.assertEqual(res["equals"]["value"], True)

    def test__equals_object_id(self):
        q = AtlasQ(f=3)
        t = AtlasTransform(q.query, AtlasIndex("test"))
        id = ObjectId("5e45de3dd2bfea029b68cce2")
        res = t._equals("field", id)
        self.assertIn("equals", res)
        self.assertIn("path", res["equals"])
        self.assertIn("value", res["equals"])
        self.assertEqual(res["equals"]["path"], "field")
        self.assertEqual(res["equals"]["value"], id)

    def test_cast_to_object_id(self):
        id = "5e45de3dd2bfea029b68cce2"
        res = AtlasTransform._cast_to_object_id(id)
        self.assertIsInstance(res, ObjectId)

        id = ObjectId("5e45de3dd2bfea029b68cce2")
        res = AtlasTransform._cast_to_object_id(id)
        self.assertIsInstance(res, ObjectId)

        id = 3
        with self.assertRaises(TypeError):
            AtlasTransform._cast_to_object_id(id)

        id = [ObjectId("5e45de3dd2bfea029b68cce2"), 3]
        with self.assertRaises(TypeError):
            AtlasTransform._cast_to_object_id(id)

        id = [ObjectId("5e45de3dd2bfea029b68cce2"), "5e45de3dd2bfea029b68cce2"]
        res = AtlasTransform._cast_to_object_id(id)
        for i in res:
            self.assertIsInstance(i, ObjectId)

    def test__equals_list_bool(self):
        q = AtlasQ(f=3)
        t = AtlasTransform(q.query, AtlasIndex("test"))
        res = t._equals("field", [True])
        self.assertIn("compound", res)
        self.assertIn("minimumShouldMatch", res["compound"])
        self.assertEqual(1, res["compound"]["minimumShouldMatch"])
        self.assertIn("should", res["compound"])
        self.assertEqual(1, len(res["compound"]["should"]))
        self.assertIn("equals", res["compound"]["should"][0])
        self.assertIn("path", res["compound"]["should"][0]["equals"])
        self.assertIn("value", res["compound"]["should"][0]["equals"])
        self.assertEqual(res["compound"]["should"][0]["equals"]["path"], "field")
        self.assertEqual(res["compound"]["should"][0]["equals"]["value"], True)

    def test(self):
        q = AtlasQ(f=3)
        t = AtlasTransform(q.query, AtlasIndex("test"))
        res = t._text("field", "aaa")
        self.assertIn("text", res)
        self.assertIn("path", res["text"])
        self.assertIn("query", res["text"])
        self.assertEqual(res["text"]["path"], "field")
        self.assertEqual(res["text"]["query"], "aaa")

    def test_none(self):
        q = AtlasQ(f=3)
        t = AtlasTransform(q.query, AtlasIndex("test"))
        with self.assertRaises(AtlasFieldError):
            t._text("field", None)

    def test__size_operator_not_supported(self):
        q = AtlasQ(f=3)
        t = AtlasTransform(q.query, AtlasIndex("test"))
        with self.assertRaises(NotImplementedError):
            t._size("field", 0, "lte")

    def test__size__not_zero(self):
        q = AtlasQ(f=3)
        t = AtlasTransform(q.query, AtlasIndex("test"))
        with self.assertRaises(NotImplementedError):
            t._size("field", 3, "ne")

    def test__size_equal_zero(self):
        q = AtlasQ(f=3)
        t = AtlasTransform(q.query, AtlasIndex("test"))
        res = t._size("field", 0, "eq")
        self.assertIn("$match", res)
        self.assertIn("field", res["$match"])
        self.assertIn("$exists", res["$match"]["field"])
        self.assertIn("$eq", res["$match"]["field"])
        self.assertEqual(res["$match"]["field"]["$exists"], True)
        self.assertCountEqual(res["$match"]["field"]["$eq"], [None, [], ""])

    def test__size_ne_zero(self):
        q = AtlasQ(f=3)
        t = AtlasTransform(q.query, AtlasIndex("test"))
        res = t._size("field", 0, "ne")
        self.assertIn("$match", res)
        self.assertIn("field", res["$match"])
        self.assertIn("$exists", res["$match"]["field"])
        self.assertIn("$ne", res["$match"]["field"])
        self.assertEqual(res["$match"]["field"]["$exists"], True)
        self.assertCountEqual(res["$match"]["field"]["$ne"], [None, [], ""])


class TestAtlasQ(TestBaseCase):
    def test_ids_in(self):
        q1 = AtlasQ(id__in=["5e45de3dd2bfea029b68cce2"])
        positive, negative, aggregations = AtlasTransform(q1.query, AtlasIndex("test")).transform()
        self.assertEqual(aggregations, [])
        self.assertEqual(negative, [])
        self.assertEqual(
            {
                "compound": {
                    "should": [
                        {
                            "equals": {
                                "path": "_id",
                                "value": ObjectId("5e45de3dd2bfea029b68cce2"),
                            }
                        }
                    ],
                    "minimumShouldMatch": 1,
                }
            },
            positive[0],
            json.dumps(positive, indent=4, default=str),
        )

    def test_size_val(self):
        q1 = AtlasQ(key__size=1)
        with self.assertRaises(NotImplementedError):
            AtlasTransform(q1.query, AtlasIndex("test")).transform()

    def test_size_ne(self):
        q1 = AtlasQ(key__size__ne=0)
        with self.assertRaises(NotImplementedError):
            AtlasTransform(q1.query, AtlasIndex("test")).transform()

    def test_size_gte(self):
        q1 = AtlasQ(key__not__size__gte=0)
        with self.assertRaises(NotImplementedError):
            AtlasTransform(q1.query, AtlasIndex("test")).transform()

    def test_size_negative(self):
        q1 = AtlasQ(key__not__size=0)
        positive, negative, aggregations = AtlasTransform(q1.query, AtlasIndex("test")).transform()
        self.assertEqual(positive, [])
        self.assertEqual(negative, [])
        self.assertEqual(
            {"$match": {"key": {"$exists": True, "$ne": [None, [], ""]}}},
            aggregations[0],
            json.dumps(aggregations, indent=4),
        )

    def test_size_positive(self):
        q1 = AtlasQ(key__size=0)
        positive, negative, aggregations = AtlasTransform(q1.query, AtlasIndex("test")).transform()
        self.assertEqual(positive, [])
        self.assertEqual(negative, [])
        self.assertEqual(
            {"$match": {"key": {"$exists": True, "$eq": [None, [], ""]}}},
            aggregations[0],
            json.dumps(aggregations, indent=4),
        )

    def test_atlas_q_not_none(self):
        q1 = AtlasQ(key__nin=["", None])
        positive, negative, aggregations = AtlasTransform(q1.query, AtlasIndex("test")).transform()
        self.assertEqual(aggregations, [])
        self.assertEqual(positive, [])
        self.assertEqual(
            [
                {"text": {"path": "key", "query": ["", None]}},
            ],
            negative,
            json.dumps(negative, indent=4),
        )

    def test_atlas_q_not_exists_3(self):
        q1 = AtlasQ(key__not__exists=True)
        positive, negative, aggregations = AtlasTransform(q1.query, AtlasIndex("test")).transform()
        self.assertEqual(positive, [])
        self.assertEqual(negative, [{"exists": {"path": "key"}}])
        self.assertEqual(
            [],
            aggregations,
        )

    def test_atlas_q_not_exists_2(self):
        q1 = AtlasQ(key__not__exists=False)
        positive, negative, aggregations = AtlasTransform(q1.query, AtlasIndex("test")).transform()
        self.assertEqual(positive, [{"exists": {"path": "key"}}])
        self.assertEqual(negative, [])
        self.assertEqual(
            [],
            aggregations,
            json.dumps(aggregations, indent=4),
        )

    def test_atlas_q_not_exists_1(self):
        q1 = AtlasQ(key__exists=False)
        positive, negative, aggregations = AtlasTransform(q1.query, AtlasIndex("test")).transform()
        self.assertEqual(positive, [])
        self.assertEqual(negative, [{"exists": {"path": "key"}}])
        self.assertEqual(
            [],
            aggregations,
            json.dumps(aggregations, indent=4),
        )

    def test_atlas_q_not_exists_0(self):
        q1 = AtlasQ(key__exists=True)
        positive, negative, aggregations = AtlasTransform(q1.query, AtlasIndex("test")).transform()
        self.assertEqual(positive, [{"exists": {"path": "key"}}])
        self.assertEqual(negative, [])
        self.assertEqual(
            [],
            aggregations,
            json.dumps(aggregations, indent=4),
        )

    def test_atlas_q_negative_equals(self):
        q1 = AtlasQ(key__ne=True)
        positive, negative, aggregations = AtlasTransform(q1.query, AtlasIndex("test")).transform()
        self.assertEqual(aggregations, [])
        self.assertEqual(positive, [])
        self.assertEqual(
            [
                {"equals": {"path": "key", "value": True}},
            ],
            negative,
            json.dumps(negative, indent=4),
        )

    def test_atlas_q_equals(self):
        q1 = AtlasQ(key=True)
        positive, negative, aggregations = AtlasTransform(q1.query, AtlasIndex("test")).transform()
        self.assertEqual([], aggregations)
        self.assertEqual([], negative)
        self.assertEqual(
            [
                {"equals": {"path": "key", "value": True}},
            ],
            positive,
            json.dumps(positive, indent=4),
        )

    def test_atlas_q_lt(self):
        date = "2021-01-01T00:00:00.000Z"
        date = datetime.datetime.strptime(date, "%Y-%m-%dT%H:%M:%S.%fZ")
        q1 = AtlasQ(last_sent_time__lt=date)
        positive, negative, aggregations = AtlasTransform(q1.query, AtlasIndex("test")).transform()
        self.assertEqual([], aggregations)
        self.assertEqual([], negative)
        self.assertEqual(
            [
                {
                    "range": {
                        "path": "last_sent_time",
                        "lt": date,
                    }
                }
            ],
            positive,
            json.dumps(positive, indent=4, default=str),
        )

    def test_atlas_q_gte(self):
        date = "2021-01-01T00:00:00.000Z"
        date = datetime.datetime.strptime(date, "%Y-%m-%dT%H:%M:%S.%fZ")
        q1 = AtlasQ(last_sent_time__gte=date)
        positive, negative, aggregations = AtlasTransform(q1.query, AtlasIndex("test")).transform()
        self.assertEqual([], aggregations)
        self.assertEqual([], negative)
        self.assertEqual(
            [
                {
                    "range": {
                        "path": "last_sent_time",
                        "gte": date,
                    }
                }
            ],
            positive,
            json.dumps(positive, indent=4, default=str),
        )

    def test_atlas_q_not_whole_word(self):
        q1 = AtlasQ(key__not__wholeword="value")
        positive, negative, aggregations = AtlasTransform(q1.query, AtlasIndex("test")).transform()
        self.assertEqual([], aggregations)
        self.assertEqual(
            [
                {"text": {"path": "key", "query": "value"}},
            ],
            negative,
        )
        self.assertEqual(
            [],
            positive,
            json.dumps(positive, indent=4),
        )

    def test_atlas_q_field_start_with_keyword(self):
        q1 = AtlasQ(key__in=["value"])
        positive, negative, aggregations = AtlasTransform(q1.query, AtlasIndex("test")).transform()
        self.assertEqual([], aggregations)
        self.assertEqual([], negative)
        self.assertEqual(
            [
                {"text": {"path": "key", "query": ["value"]}},
            ],
            positive,
            json.dumps(positive, indent=4),
        )

    def test_atlas_q_ne_embedded_document(self):
        index = AtlasIndex("test")
        index._indexed_fields = {
            "f": "embeddedDocuments",
            "f.g": "embeddedDocuments",
            "f.g.h": "string",
        }
        index.ensured = True
        q = AtlasQ(f__g__h__ne="test")
        positive, negative, aggregations = AtlasTransform(q.query, index).transform()
        self.assertEqual(
            positive,
            [
                {
                    "embeddedDocument": {
                        "path": "f",
                        "operator": {
                            "compound": {
                                "must": [
                                    {
                                        "embeddedDocument": {
                                            "path": "f.g",
                                            "operator": {
                                                "compound": {
                                                    "mustNot": [
                                                        {
                                                            "text": {
                                                                "query": "test",
                                                                "path": "f.g.h",
                                                            }
                                                        },
                                                    ]
                                                }
                                            },
                                        }
                                    },
                                ]
                            }
                        },
                    }
                }
            ],
            json.dumps(positive, indent=4),
        )
        self.assertEqual(negative, [])
        self.assertEqual(aggregations, [])

    def test_atlas_q_embedded_document(self):
        q = AtlasQ(f__g__h="test")
        index = AtlasIndex("test")
        index._indexed_fields = {
            "f": "embeddedDocuments",
            "f.g": "embeddedDocuments",
            "f.g.h": "string",
        }
        index.ensured = True
        positive, negative, aggregations = AtlasTransform(q.query, index).transform()
        self.assertEqual(
            positive,
            [
                {
                    "embeddedDocument": {
                        "path": "f",
                        "operator": {
                            "compound": {
                                "must": [
                                    {
                                        "embeddedDocument": {
                                            "path": "f.g",
                                            "operator": {
                                                "compound": {
                                                    "must": [
                                                        {
                                                            "text": {
                                                                "query": "test",
                                                                "path": "f.g.h",
                                                            }
                                                        }
                                                    ]
                                                }
                                            },
                                        }
                                    }
                                ]
                            }
                        },
                    }
                }
            ],
            json.dumps(positive, indent=4),
        )
        self.assertEqual(negative, [])
        self.assertEqual(aggregations, [])

    def test_atlas_q_nin(self):
        q1 = AtlasQ(key__nin=["value", "value2"])
        positive, negative, aggregations = AtlasTransform(q1.query, AtlasIndex("test")).transform()
        self.assertEqual(aggregations, [])
        self.assertEqual(positive, [])
        self.assertEqual(
            [
                {"text": {"path": "key", "query": ["value", "value2"]}},
            ],
            negative,
            json.dumps(negative, indent=4),
        )

    def test_atlas_q_negative(self):
        q1 = AtlasQ(key__ne="value")
        positive, negative, aggregations = AtlasTransform(q1.query, AtlasIndex("test")).transform()
        self.assertEqual(aggregations, [])
        self.assertEqual(positive, [])

        self.assertEqual(
            [
                {"text": {"path": "key", "query": "value"}},
            ],
            negative,
            json.dumps(negative, indent=4),
        )

    def test_atlas_q(self):
        q1 = AtlasQ(key="value", key2="value2")
        positive, negative, aggregations = AtlasTransform(q1.query, AtlasIndex("test")).transform()
        self.assertEqual([], aggregations)
        self.assertEqual([], negative)
        self.assertEqual(
            [
                {"text": {"path": "key", "query": "value"}},
                {"text": {"path": "key2", "query": "value2"}},
            ],
            positive,
            json.dumps(positive, indent=4),
        )



import datetime
import logging
from typing import Any, Dict, List, Tuple, Union

from atlasq.queryset.exceptions import AtlasFieldError, AtlasIndexFieldError
from atlasq.queryset.index import AtlasIndex, AtlasIndexType
from bson import ObjectId
from mongoengine import QuerySet

logger = logging.getLogger(__name__)


class AtlasTransform:

    id_keywords = [
        "pk",
        "id",
        "_id",
    ]

    keywords = [
        "ne",
        "lt",
        "gt",
        "lte",
        "gte",
        "in",
        "nin",
        "all",
        "size",
        "exists",
        "exact",
        "iexact",
        "contains",
        "icontains",
        "startswith",
        "istartswith",
        "iwholeword",
        "wholeword",
        "not",
        "mod",
        "regex",
        "iregex",
        "match",
        "is",
    ]
    negative_keywords = ["ne", "nin", "not"]
    exists_keywords = ["exists"]
    range_keywords = ["gt", "gte", "lt", "lte"]
    equals_keywords = []
    equals_type_supported = (bool, ObjectId, int, datetime.datetime)
    text_keywords = [
        "contains",
        "icontains",
        "iwholeword",
        "wholeword",
        "exact",
        "iexact",
        "eq",
    ]
    all_keywords = ["all"]
    regex_keywords = ["regex", "iregex"]
    size_keywords = ["size"]
    not_converted = [
        "istartswith",
        "startswith",
        "contains",
        "icontains",
        "mod",
        "match",
    ]

    def __init__(self, atlas_query, atlas_index: AtlasIndex):
        self.atlas_query = atlas_query
        self.atlas_index = atlas_index

    def _all(self, path: str, values: List[str]):
        filters = []
        for value in values:
            filters.append(self._auto_convert_type_to_keyword(path, value))
        return {"compound": {"must": filters}}

    def _regex(self, path: str, value: str):
        return {"regex": {"query": value, "path": path}}

    def _embedded_document(self, path: str, content: Dict, positive: bool):
        operator = "must" if positive else "mustNot"
        return {
            "embeddedDocument": {
                "path": path,
                "operator": {"compound": {operator: [content]}},
            }
        }

    def _convert_to_embedded_document(self, path: List[str], operator: Dict, positive: bool, start: str = ""):
        element = path.pop(0)
        partial_path = f"{start}.{element}" if start else element
        if not self.atlas_index.ensured:
            return operator
        if self.atlas_index.get_type_from_keyword(partial_path) != AtlasIndexType.EMBEDDED_DOCUMENT.value:
            return operator

        if not path:
            return operator

        new_operator = self._convert_to_embedded_document(path, operator, start=partial_path, positive=positive)
        return self._embedded_document(
            partial_path,
            new_operator,
            True if operator != new_operator else positive,  # this cover the case of multiple embeddedDocument,
            # where only the last one must be set to negative
        )

    def _exists(self, path: str) -> Dict:
        return {"exists": {"path": path}}

    def _range(self, path: str, value: Union[int, datetime.datetime], keywords: List[str]) -> Dict:
        for keyword in keywords:
            if keyword not in self.range_keywords:
                raise AtlasFieldError(f"Range search for {path} must be {self.range_keywords}, not {keyword}")
        if isinstance(value, datetime.datetime):
            value = value.replace(microsecond=0)
        elif isinstance(value, int):
            pass
        else:
            raise AtlasFieldError(f"Range search for {path} must have a value of datetime or integer")
        return {"range": {"path": path, **{keyword: value for keyword in keywords}}}

    def _single_equals(self, path: str, value: Union[ObjectId, bool]):
        if not isinstance(value, self.equals_type_supported):
            raise AtlasFieldError(f"Text search for equals on {path=} cannot be {value}, must be ObjectId or bool")
        return {
            "equals": {
                "path": path,
                "value": value,
            }
        }

    def _equals(self, path: str, value: Union[List[Union[ObjectId, bool]], ObjectId, bool]) -> Dict:
        if isinstance(value, list):

            values = value
            if not values:
                raise AtlasFieldError(f"Text search for equals on {path=} cannot be empty")
            base = {"compound": {"should": [], "minimumShouldMatch": 1}}
            for value in values:
                base["compound"]["should"].append(self._single_equals(path, value))
            return base
        return self._single_equals(path, value)

    def _text(self, path: str, value: Any) -> Dict:
        if not value:
            raise AtlasFieldError(f"Text search for {path} cannot be {value}")
        return {
            "text": {"query": value, "path": path},
        }

    def _size(self, path: str, value: int, operator: str) -> Dict:
        if not isinstance(value, int):
            raise AtlasFieldError(f"Size search for {path} must be an int")
        if value != 0:
            raise NotImplementedError(f"Size search for {path} must be 0")
        if operator not in ["eq", "ne"]:
            raise NotImplementedError(f"Size search for {path} must be eq or ne")
        return {
            "$match": {
                path: {
                    "$exists": True,
                    f"${operator}": [None, [], ""],
                }
            }
        }

    def _ensure_path_is_indexed(self, path: List[str]) -> None:
        start = ""
        for element in path:
            partial_path = f"{start}.{element}" if start else element

            if not self.atlas_index.ensure_keyword_is_indexed(partial_path):
                raise AtlasIndexFieldError(f"The keyword {partial_path} is not indexed in {self.atlas_index.index}")
            start = partial_path

    @staticmethod
    def _cast_to_object_id(value: Union[str, ObjectId, List[Union[str, ObjectId]]]) -> Union[ObjectId, List[ObjectId]]:
        if isinstance(value, str):
            value = ObjectId(value)
        elif isinstance(value, list):
            for j in range(len(value)):  # pylint: disable=consider-using-enumerate
                if isinstance(value[j], str):
                    value[j] = ObjectId(value[j])
                elif isinstance(value[j], ObjectId):
                    pass
                else:
                    raise TypeError(f"Wrong type {type(value[j])} for id field")
        elif isinstance(value, ObjectId):
            pass
        else:
            raise TypeError(f"Wrong type {type(value)} for id field")
        return value

    def _auto_convert_type_to_keyword(self, path: str, value: Any) -> Dict:
        if isinstance(value, list):
            value_to_check = value[0]
            if not all(isinstance(x, value_to_check.__class__) or x is None for x in value):
                raise TypeError(f"All elements of a list should have the same type: {value}")
        elif isinstance(value, dict):
            raise TypeError(f"It is not possible to have a dictionary as a value: {value}")
        else:
            value_to_check = value
        if isinstance(value_to_check, self.equals_type_supported):
            obj = self._equals(path, value)
        else:
            obj = self._text(path, value)
        return obj

    def transform(self) -> Tuple[List[Dict], List[Dict], List[Dict]]:
        other_aggregations = []
        affirmative = []
        negative = []

        for key, value in self.atlas_query.items():
            # if to_go is positive, we add the element in the positive list
            # if to_go is negative, we add the element in the negative list
            to_go = 1
            if isinstance(value, QuerySet):
                logger.debug("Casting queryset to list, otherwise the aggregation will fail")
                value = list(value)
            key_parts = key.split("__")
            obj = None
            path = ""
            for i, keyword in enumerate(key_parts):

                if keyword in self.id_keywords:
                    keyword = "_id"
                    key_parts[i] = keyword
                    value = self._cast_to_object_id(value)
                if keyword not in self.keywords:
                    continue
                # the key_part is made of field__subfield__keywords
                # meaning that the first time that we find a keyword, we have the entire path
                if not path:
                    path = ".".join(key_parts[:i])

                if keyword in self.not_converted:
                    raise NotImplementedError(f"Keyword {keyword} not implemented yet")
                if keyword in self.negative_keywords:
                    to_go *= -1

                if keyword in self.size_keywords:
                    # it must the last keyword, otherwise we do not support it
                    if i != len(key_parts) - 1:
                        raise NotImplementedError(f"Keyword {keyword} not implemented yet")
                    other_aggregations.append(self._size(path, value, "eq" if to_go == 1 else "ne"))
                    break
                if keyword in self.exists_keywords:
                    if value is False:
                        to_go *= -1
                    obj = self._exists(path)
                    break

                if keyword in self.range_keywords:
                    obj = self._range(path, value, [keyword])
                    break
                if keyword in self.equals_keywords:
                    obj = self._equals(path, value)
                    break
                if keyword in self.text_keywords:
                    obj = self._text(path, value)
                    break
                if keyword in self.regex_keywords:
                    obj = self._regex(path, value)
                    break
                if keyword in self.all_keywords:
                    obj = self._all(path, value)
                    break
            else:
                if not path:
                    path = ".".join(key_parts)
                obj = self._auto_convert_type_to_keyword(path, value)

            if obj:
                if self.atlas_index.ensured:
                    self._ensure_path_is_indexed(path.split("."))
                # we are wrapping the result to an embedded document
                converted = self._convert_to_embedded_document(path.split("."), obj, positive=to_go == 1)
                if obj != converted:
                    # we have an embedded object
                    # the mustNot is done inside the embedded document clause
                    affirmative = self.merge_embedded_documents(converted, affirmative)
                else:
                    if to_go == 1:
                        affirmative.append(converted)
                    else:
                        negative.append(converted)
        if other_aggregations:
            logger.warning(
                "CARE! You are generating a query that uses other aggregations other than text search!"
                "Meaning that the query will be slower."
                f" Aggregations generated are {other_aggregations}"
            )
        return affirmative, negative, other_aggregations

    @staticmethod
    def merge_embedded_documents(obj: Dict, list_of_obj: List[Dict]) -> List[Dict]:
        list_of_obj = list(list_of_obj)  # I hate function that change stuff in place
        assert "embeddedDocument" in obj
        assert "path" in obj["embeddedDocument"]
        assert "operator" in obj["embeddedDocument"]
        assert "compound" in obj["embeddedDocument"]["operator"]
        # path that we want merge
        path = obj["embeddedDocument"]["path"]
        keys = list(obj["embeddedDocument"]["operator"]["compound"].keys())
        assert len(keys) == 1
        operator = keys[0]  # values could be (must, mustNot)
        # the actual query
        content = obj["embeddedDocument"]["operator"]["compound"][operator]
        for already_present_obj in list_of_obj:
            # we have added an object that is not actually an embedded object, nothing to do
            if "embeddedDocument" not in already_present_obj:
                continue
            # we check for a correspondence
            if path == already_present_obj["embeddedDocument"]["path"]:
                # we merge the objects
                already_present_obj["embeddedDocument"]["operator"]["compound"].setdefault(operator, []).extend(content)
                # we can exit since we are sure that it can be only 1 hit for path
                # if this method is called at every embedded object
                break
        # otherwise we just add the object if no hit has been found
        else:
            list_of_obj.append(obj)
        return list_of_obj



import copy
import logging
import time
from typing import Any, Dict, List, Tuple

from atlasq.queryset.exceptions import AtlasIndexError
from atlasq.queryset.index import AtlasIndex
from atlasq.queryset.node import AtlasQ
from mongoengine import Q, QuerySet
from pymongo.command_cursor import CommandCursor

logger = logging.getLogger(__name__)


def clock(func):
    def clocked(self, *args, **kwargs):
        start_time = time.perf_counter()
        result = func(self, *args, **kwargs)
        elapsed = time.perf_counter() - start_time
        floor = f"{elapsed:0.3f}"
        logger.info(f"{floor} - {result}")
        return result

    return clocked


# pylint: disable=too-many-instance-attributes
class AtlasQuerySet(QuerySet):
    def _clone_into(self, new_qs):
        copy_props = (
            "index",
            "_aggrs_query",
            "_search_result",
            "_count",
            "_return_objects",
            "_other_aggregations",
        )
        qs = super()._clone_into(new_qs)
        for prop in copy_props:
            val = getattr(self, prop)
            setattr(new_qs, prop, copy.copy(val))
        return qs

    def __init__(self, document, collection):
        super().__init__(document, collection)
        self._query_obj = AtlasQ()
        self.index: AtlasIndex = None

        self._aggrs_query: List[Dict[str, Any]] = None
        self._search_result: CommandCursor = None
        self._count: bool = False
        self._return_objects: bool = True
        self._other_aggregations: List[Dict] = []

    # pylint: disable=too-many-arguments
    def upload_index(
        self,
        json_index: Dict,
        user: str,
        password: str,
        group_id: str,
        cluster_name: str,
    ):
        db_name = self._document._get_db().name  # pylint: disable=protected-access
        collection_name = self._document._get_collection_name()  # pylint: disable=protected-access
        if "collectionName" not in json_index:
            json_index["collectionName"] = collection_name
        if "database" not in json_index:
            json_index["database"] = db_name
        if "name" not in json_index:
            json_index["name"] = self.index._index  # pylint: disable=protected-access
        logger.info(f"Sending {json_index} to create new index")
        return self.index.upload_index(json_index, user, password, group_id, cluster_name)

    def ensure_index(self, user: str, password: str, group_id: str, cluster_name: str):
        db_name = self._document._get_db().name  # pylint: disable=protected-access
        collection_name = self._document._get_collection_name()  # pylint: disable=protected-access
        return self.index.ensure_index_exists(user, password, group_id, cluster_name, db_name, collection_name)

    def __iter__(self):
        if not self._return_objects:
            return iter(self._cursor)
        return super().__iter__()

    def delete(self, write_concern=None, _from_doc_delete=False, cascade_refs=None):
        # we need to get the mongoengine query, the fastest way it to just call _cursor
        assert self._cursor is not None
        assert self._query is not None
        return super().delete(
            write_concern=write_concern,
            _from_doc_delete=_from_doc_delete,
            cascade_refs=cascade_refs,
        )

    @property
    @clock
    def _aggrs(self):
        # corresponding of _query for us
        if self._aggrs_query is None:
            self._aggrs_query = self._query_obj.to_query(self._document)
            if self._aggrs_query:
                if self._count:
                    self._aggrs_query[0]["$search"]["count"] = {"type": "total"}
            self._aggrs_query += self._get_projections()
            self._aggrs_query += self._other_aggregations
        return self._aggrs_query

    @property
    def _cursor(self):
        if not self._search_result:
            self._search_result = self.__collection_aggregate(self._aggrs)
        if not self._return_objects:
            self._cursor_obj = self._search_result  # pylint: disable=attribute-defined-outside-init
        cursor = super()._cursor
        return cursor

    def order_by(self, *keys):
        if not keys:
            return self
        qs: AtlasQuerySet = self.clone()
        order_by: List[Tuple[str, int]] = qs._get_order_by(keys)  # pylint: disable=protected-access
        aggregation = {"$sort": dict(order_by)}
        qs._other_aggregations.append(aggregation)  # pylint: disable=protected-access
        return qs

    def __getitem__(self, key):
        if isinstance(key, int):
            from mongoengine.queryset.base import BaseQuerySet

            qs = self.clone()
            qs._limit = 1
            return BaseQuerySet.__getitem__(qs, key)
        return super().__getitem__(key)

    @property
    def _query(self):
        if not self._search_result:
            return None
        start = self._skip
        end = self._limit
        # unfortunately here we have to actually run the query to get the objects
        # I do not see other way to do this atm
        ids: List[str] = []
        for i, obj in enumerate(self._search_result):
            if end is not None and i >= end:
                break
            if start is not None and i < start:
                continue
            if obj:
                ids.append(obj["_id"])
        self._query_obj = Q(id__in=ids)
        logger.debug(self._query_obj.to_query(self._document))
        return super()._query

    def __collection_aggregate(self, final_pipeline, **kwargs):
        collection = self._collection
        if self._read_preference is not None or self._read_concern is not None:
            collection = self._collection.with_options(read_preference=self._read_preference, read_concern=self._read_concern)
        logger.info(final_pipeline)
        return collection.aggregate(final_pipeline, cursor={}, **kwargs)

    def aggregate(self, pipeline, **kwargs):  # pylint: disable=arguments-differ,unused-argument
        self._return_objects = False
        if isinstance(pipeline, dict):
            pipeline = [pipeline]

        final_pipeline = self._aggrs + pipeline
        return self.__collection_aggregate(final_pipeline)

    def __call__(self, q_obj=None, **query):
        if self.index is None:
            raise AtlasIndexError("Index is not set")

        q = AtlasQ(**query)
        if q_obj is not None:
            q &= q_obj
        logger.debug(q)
        qs = super().__call__(q)
        return qs

    def _get_projections(self) -> List[Dict[str, Any]]:
        if self._count:
            if self._query_obj:
                return [{"$project": {"meta": "$$SEARCH_META"}}]
            return [{"$count": "count"}]
        loaded_fields = self._loaded_fields.as_dict()
        logger.debug(loaded_fields)
        if loaded_fields:
            return [{"$project": loaded_fields}]
        return []

    def count(self, with_limit_and_skip=False):  # pylint: disable=unused-argument
        self._count = True  # pylint: disable=protected-access
        cursor = self.__collection_aggregate(self._aggrs)  # pylint: disable=protected-access
        try:
            count = next(cursor)
        except StopIteration:
            self._len = 0  # pylint: disable=attribute-defined-outside-init
        else:
            logger.debug(count)
            if self._query_obj:
                self._len = count["meta"]["count"]["total"]  # pylint: disable=attribute-defined-outside-init
            else:
                self._len = count["count"]  # pylint: disable=attribute-defined-outside-init
        logger.debug(self._len)
        return self._len

    def limit(self, n):
        qs = self.clone()
        qs._limit = n  # pylint: disable=protected-access
        qs._other_aggregations.append({"$limit": n})  # pylint: disable=protected-access
        return qs
